import asyncio
import datetime
import logging
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from functools import singledispatchmethod
from importlib import metadata

import httpx

from wedne.commands import CommandSchema
from wedne.utils import Backoffer


@dataclass
class Watcher:
    endpoint: str
    delay: datetime.timedelta
    social_media_id: int
    command_processor: Callable[[CommandSchema | None], Awaitable[None]]

    def __post_init__(self) -> None:
        self.backoffer = Backoffer()

    async def __call__(self) -> None:
        version = metadata.version("wedne")
        client = httpx.AsyncClient(
            headers={"user-agent": f"wedne.client/{version}"},
            timeout=10,
        )
        try:
            await self._watch(client)
        finally:
            await client.aclose()

    async def _watch(self, client: httpx.AsyncClient) -> None:
        data = {
            "social_media_id": self.social_media_id,
        }
        while True:
            try:
                response = await client.post(
                    self.endpoint,
                    json=data,
                )
                await self.process(response)
                await asyncio.sleep(self.delay.total_seconds())
            except httpx.ConnectError as exc:
                await self.process(exc)
                await asyncio.sleep(self.backoffer.failed())

    @singledispatchmethod
    async def process(self, response: httpx.Response) -> None:
        logging.info("response: %s", response.text)
        if response.is_server_error:
            logging.warning("received a server error")
            await asyncio.sleep(self.backoffer.failed())
            return

        if response.is_client_error:
            logging.warning(
                "received a client error: probably you should update a package.",
            )
            await asyncio.sleep(self.backoffer.failed())
            return

        self.backoffer.succeeded()

        raw_command = response.json()
        command = CommandSchema.parse_obj(raw_command) if raw_command else None
        await self.command_processor(command)

    @process.register
    async def _(self, exception: httpx.ConnectError) -> None:
        logging.error("connect error: %s", str(exception))



import asyncio
from importlib import metadata

import aiorun
import fire

from wedne.client.consts import API_HASH, API_ID, WATCHING_DELAY
from wedne.client.telegram import TelegramTowerBuilder
from wedne.client.watcher import Watcher


async def main(
    endpoint: str,
    destroying: bool,
    chat_id: int,
) -> None:
    telegram_tower_builder = TelegramTowerBuilder(
        "wedne",
        API_ID,
        API_HASH,
        chat_id,
    )
    await telegram_tower_builder.start()
    monitor_task = asyncio.create_task(telegram_tower_builder.monitor(destroying))
    watcher = Watcher(
        endpoint,
        delay=WATCHING_DELAY,
        social_media_id=await telegram_tower_builder.who_am_i(),
        command_processor=telegram_tower_builder.process_command,
    )
    watcher_task = asyncio.create_task(watcher())
    await asyncio.wait(
        [
            monitor_task,
            watcher_task,
        ],
    )


def sync_main(
    endpoint: str,
    destroying: bool = False,
    chat_id: int = -1001481658345,  # "Вастрик.Бар"
) -> None:
    print(metadata.version("wedne"))
    aiorun.run(main(endpoint, destroying, chat_id), stop_on_unhandled_errors=True)


if __name__ == "__main__":
    fire.core.Display = lambda lines, out: out.write(  # type: ignore
        "\n".join(lines) + "\n",
    )
    fire.Fire(sync_main, name="wedne.client")



from datetime import datetime, timedelta

import fastapi
import pydantic
import pytz

from wedne.server.db import VisitDAO
from wedne.server.settings import settings
from wedne.utils import distinct_on

router = fastapi.APIRouter()


class StatSchema(pydantic.BaseModel):
    duration: timedelta
    visits: int
    unique_visits: int
    tower: str


@router.get("/")
def get_statistics(
    dao: VisitDAO = fastapi.Depends(),
) -> StatSchema:
    duration = timedelta(minutes=settings.minutes_of_last_activity)
    active = dao.get_from(
        datetime.now(pytz.utc) - duration,
    )
    unique = distinct_on(active, lambda vis: vis.social_media_id)
    return StatSchema(
        duration=duration,
        visits=len(active),
        unique_visits=len(unique),
        tower=settings.tower,
    )



from pathlib import Path

from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    minutes_to_order: int
    seconds_of_delay: int
    minutes_of_last_activity: int
    app_database: Path
    job_queue_database: Path
    tower: str = "ITSWEDNESDAYMYDUDES"
    host: str
    port: int
    reload: bool = False
    model_config = SettingsConfigDict(env_file=".env", env_prefix="wedne_")


settings = Settings()  # type: ignore



from wedne.utils import Backoffer, distinct_on


def test_distinct_on():
    unique_roots = distinct_on([4, 5, 9], key=lambda x: int(x**0.5))
    assert sorted(unique_roots) == [5, 9]
    unique_roots = distinct_on([5, 4, 9], key=lambda x: int(x**0.5))
    assert sorted(unique_roots) == [4, 9]


def test_backoffer_smoke():
    backoffer = Backoffer(base=2, max_value=100500)
    assert backoffer.failed() == 1
    assert backoffer.failed() == 2
    assert backoffer.failed() == 4
    backoffer.succeeded()
    assert backoffer.failed() == 1


def test_backoffer_with_max_value():
    backoffer = Backoffer(base=10, max_value=69)
    backoffer.failed()
    backoffer.failed()
    assert backoffer.failed() == 69



import logging

import fire
import uvicorn
from huey.consumer_options import ConsumerConfig

from wedne.server.settings import settings
from wedne.server.tasks import job_queue


class Main:
    """
    Server for tower building

    Everything is configured with environment variables or .env file.
    """

    def run_api(self) -> None:
        """
        Server entrypoint
        """
        uvicorn.run(
            "wedne.server.api:app",
            host=settings.host,
            port=settings.port,
            reload=settings.reload,
        )

    def consume_tasks(self) -> None:
        """
        Run Huey Consumer

        It's a task executor which is used for almost everything.
        """
        config = ConsumerConfig(
            workers=1,
            periodic=True,
            max_delay=2.0,
        )
        logger = logging.getLogger("huey")
        config.setup_logger(logger)

        consumer = job_queue.create_consumer(**config.values)
        consumer.run()


if __name__ == "__main__":
    fire.core.Display = lambda lines, out: out.write(  # type: ignore
        "\n".join(lines) + "\n",
    )
    fire.Fire(Main(), name="wedne.server")



from collections.abc import Callable, Hashable, Iterable, Sequence
from typing import TypeVar

T_ = TypeVar("T_")


def distinct_on(seq: Iterable[T_], key: Callable[[T_], Hashable]) -> Sequence[T_]:
    unique = {key(element): element for element in seq}
    return list(unique.values())


class Backoffer:
    def __init__(self, base: float = 2, max_value: float = 32):
        self.base = base
        self.max_value = max_value
        self.counter = 0

    def failed(self) -> float:
        result = min(self.max_value, self.base**self.counter)
        self.counter += 1
        return result

    def succeeded(self) -> None:
        self.counter = 0



import fire

from wedne.client.__main__ import sync_main as client_main

if __name__ == "__main__":
    fire.core.Display = lambda lines, out: out.write(  # type: ignore
        "\n".join(lines) + "\n",
    )
    fire.Fire(client_main, name="wedne")



import datetime

import pytest

from wedne.commands import CommandSchema


def test_fail_command_validation_on_word():
    with pytest.raises(ValueError):
        _ = CommandSchema(letter="long word", after=None, when=datetime.datetime.min)



import datetime

# This is official tdesktop snap's API key. It's obfuscated because I'm not
# sure Telegram would be happy to see it here
API_ID = 123456 + 487879
API_HASH = "f".join(["c9ca14", "d1c48680", "73d4", "12d414b425d"])[::-1]

WATCHING_DELAY = datetime.timedelta(seconds=2)
WRITE_FIRST_DELAY = datetime.timedelta(seconds=3.77)

DESTROYING_TEXT = "ПУК"



import json
from collections.abc import Hashable
from datetime import datetime
from functools import partial
from typing import Any

import peewee
from playhouse.sqlite_ext import JSONField

from wedne.server.settings import settings

database = peewee.SqliteDatabase(
    settings.app_database,
    autoconnect=False,
    pragmas={"journal_mode": "wal"},
)


class AutoFieldType(int, peewee.AutoField):
    pass


class JSONFieldType(dict[Hashable, Any], JSONField):
    pass


class VisitModel(peewee.Model):
    id: AutoFieldType = peewee.AutoField(primary_key=True)  # type: ignore
    when: datetime = peewee.DateTimeField(index=True)  # type: ignore
    social_media_id: int = peewee.BigIntegerField()  # type: ignore
    command: JSONFieldType = JSONField(
        json_dumps=partial(json.dumps, default=str),
        null=True,
    )  # type: ignore

    class Meta:
        database = database


class VisitDAO:
    def create(self, time: datetime, social_media_id: int) -> VisitModel:
        with database:
            return VisitModel.create(when=time, social_media_id=social_media_id)

    def update_command(self, id: int, command: dict[str, Any]) -> None:
        with database:
            VisitModel.update(command=command).where(VisitModel.id == id).execute()

    def get_from(self, from_: datetime) -> list[VisitModel]:
        with database:
            return list(VisitModel.select().where(VisitModel.when >= from_))

    def get_command(
        self,
        from_: datetime,
        social_media_id: int,
    ) -> None | dict[str, Any]:
        with database:
            visit: VisitModel | None = (
                VisitModel.select(VisitModel.command)
                .where(VisitModel.social_media_id == social_media_id)
                .where(VisitModel.when >= from_)
                .where(VisitModel.command.is_null(False))
                .get_or_none()
            )
        return None if visit is None else visit.command  # type: ignore

    def get_commanded(self, from_: datetime) -> list[VisitModel]:
        with database:
            return list(
                VisitModel.select()
                .where(VisitModel.when >= from_)
                .where(VisitModel.command.is_null(False)),
            )


with database:
    database.create_tables([VisitModel])



import asyncio

from wedne.commands import CommandSchema


class SharedCommand:
    def __init__(self):
        self._lock = asyncio.Lock()
        self._command: CommandSchema | None = None

    async def write(self, command: CommandSchema) -> None:
        async with self._lock:
            self._command = command

    async def clear(self) -> None:
        async with self._lock:
            self._command = None

    async def read(self) -> CommandSchema | None:
        async with self._lock:
            return self._command
